#ifndef READWRITELOCK_H_INCLUDED
#define READWRITELOCK_H_INCLUDED

#pragma once

// KReadWriteLock is a read/write mutex.
//
// Any number of readers can acquire the mutex for read (LockRead()).
// Only single writer can acquire the mutex for write (LockWrite()).
// No other writers or readers can hold the mutex while writer uses it.
//
// KReadWriteLock does not prefer neither writers nor readers.
// I.e. if both writers and readers are waiting for the lock,
// any of them can grab the lock when it is released by previous owner.
//
// KReadWriteLock is recursive, but not upgradable. This means that
//
// if you have a read lock, you can safely acquire another read lock.
// if you have a read lock, you CANNOT acquire a write lock (deadlock!).
//
// if you have a write lock, you can safely acquire another write lock.
// if you have a write lock, you CANNOT acquire a read lock (deadlock!).
//
class KReadWriteLock
{
    CMutex m_Access;
    int    m_Readers;
    CMutex m_WriterMutex;
    CEvent m_CanRead;
    CEvent m_CanWrite;

    KReadWriteLock( KReadWriteLock const& ); // not implemented
    void operator=( KReadWriteLock const& ); // not implemented

public:
    KReadWriteLock();
    
    bool LockRead( DWORD Timeout = INFINITE );
    bool LockWrite( DWORD Timeout = INFINITE );
    bool UnlockRead( DWORD Timeout = INFINITE );
    bool UnlockWrite( DWORD Timeout = INFINITE );
};


#pragma warning( disable: 4786 )
#include <map>

// KReadWriteLockEx is more sofisticated read/write mutex.
// It is recursive and upgradable.
//
// That is, a thread can acquire and release read and write locks in arbitrary order
// The only rule is that every successfully acquired lock must be eventually released
//
// E.g. LockRead(), LockWrite(), UnlockRead(), UnlockWrite() is allowed sequence.
//
//
// KReadWriteLockEx prefers writers over readers. E.g., when a writer
// is waiting for the lock, no new readers can acquire the lock. Existing
// readers may continue to use the lock. This policy prevents so-called
// "writer starvation" when writers never get access to the lock because of
// too much readers.
//
class KReadWriteLockEx
{
    struct KThreadInfo
    {
        int ReadCount;
        int WriteCount;

        bool IsReader() { return ReadCount > 0; };
        bool IsWriter() { return WriteCount > 0; };
    };

    typedef std::map<DWORD, KThreadInfo> KThreadMap;
    KThreadMap m_Threads;

    // Number of reader threads; 
    // Note: writer threads are not reader threads, even if they have read locks
    int m_TotalReaderThreads; 
    int m_TotalWriterThreads; // current writer + waiting writers

    CMutex m_Access;
    CMutex m_WriterMutex;
    CEvent m_CanRead;
    CEvent m_CanWrite;

    void IncReaders()
        {
            ++m_TotalReaderThreads;
            ResetEvent(m_CanWrite);
        };

    void DecReaders()
        {
            if (--m_TotalReaderThreads == 0)
                SetEvent(m_CanWrite);
        };

    void IncWriters()
        {
            ++m_TotalWriterThreads;
            ResetEvent(m_CanRead);
        };

    void DecWriters()
        {
            if (--m_TotalWriterThreads == 0)
                SetEvent(m_CanRead);
        };

    KReadWriteLockEx( KReadWriteLockEx const& ); // not implemented
    void operator=( KReadWriteLockEx const& ); // not implemented

public:
    KReadWriteLockEx();
    
    bool LockRead( DWORD Timeout = INFINITE );
    bool LockWrite( DWORD Timeout = INFINITE );
    bool UnlockRead( DWORD Timeout = INFINITE );
    bool UnlockWrite( DWORD Timeout = INFINITE );

};

// Template class below automatically acquires a read lock in constructor
// and releases it in destructor. It is similar to MFC's CSingleLock class.
// NOTE: unlike CSingleLock, KReadLockTempl will lock the resource by default
// (bInitialLock parameter defaults to true).
//
template <class T>
class KReadLockTempl
{
    bool m_bLocked;
    T&   m_Lock;

    KReadLockTempl( KReadLockTempl const& ); // not implemented
    void operator=( KReadLockTempl const& ); // not implemented


public:
    KReadLockTempl(T& Lock, bool bInitialLock = true)
        :
        m_Lock(Lock),
        m_bLocked(false)
        {
            if (bInitialLock)
            {
                VERIFY(m_Lock.LockRead());
                m_bLocked = true;
            }
        };

    ~KReadLockTempl()
        {
            if (m_bLocked)
                VERIFY(m_Lock.UnlockRead());
        };

    bool Lock(DWORD Timeout = INFINITE)
        {
            ASSERT(!m_bLocked);
            m_bLocked = m_Lock.LockRead(Timeout);
            return m_bLocked;
        };

    bool Unlock(DWORD Timeout = INFINITE)
        {
            ASSERT(m_bLocked);
            m_bLocked = !m_Lock.UnlockRead(Timeout);
            return !m_bLocked;
        }
};

// Template class below automatically acquires a write lock in constructor
// and releases it in destructor. It is similar to MFC's CSingleLock class.
// NOTE: unlike CSingleLock, KWriteLockTempl will lock the resource by default
// (bInitialLock parameter defaults to true).
//
template <class T>
class KWriteLockTempl
{
    bool m_bLocked;
    T&   m_Lock;

    KWriteLockTempl( KWriteLockTempl const& ); // not implemented
    void operator=( KWriteLockTempl const& ); // not implemented


public:
    KWriteLockTempl(T& Lock, bool bInitialLock = true)
        :
        m_Lock(Lock),
        m_bLocked(false)
        {
            if (bInitialLock)
            {
                VERIFY(m_Lock.LockWrite());
                m_bLocked = true;
            }
        };

    ~KWriteLockTempl()
        {
            if (m_bLocked)
                VERIFY(m_Lock.UnlockWrite());
        };

    bool Lock(DWORD Timeout = INFINITE)
        {
            ASSERT(!m_bLocked);
            m_bLocked = m_Lock.LockWrite(Timeout);
            return m_bLocked;
        };

    bool Unlock(DWORD Timeout = INFINITE)
        {
            ASSERT(m_bLocked);
            m_bLocked = !m_Lock.UnlockWrite(Timeout);
            return !m_bLocked;
        }
};

typedef KReadLockTempl<KReadWriteLock> KReadLock;
typedef KWriteLockTempl<KReadWriteLock> KWriteLock;
typedef KReadLockTempl<KReadWriteLockEx> KReadLockEx;
typedef KWriteLockTempl<KReadWriteLockEx> KWriteLockEx;


#endif //READWRITELOCK_H_INCLUDED
